import re
import pathlib
import datetime

import requests
from dateutil.parser import parse as parse_date
from w3lib import html


def safe_open(filename: str, mode='r'):
    file = pathlib.Path(filename)

    if 'w' in mode and not file.parent.exists():
        file.parent.mkdir(parents=True)

    if filename.endswith('.gz'):
        import gzip
        return gzip.open(filename, mode=mode)

    return file.open(mode=mode)


def check_email(text: str):
    res = re.findall(r'([^\s]+?@.+)\.', text)
    return res[0] if res else None


def replace_entities(text: str):
    """
        multi level entites:
            - eg. PMID=33866717:  '&amp;lt;10' => '&lt;10' => '<10'
    """
    raw_text = text
    n = 5
    while n > 0:
        if re.search(r'&\w{2,7};', text):
            text = html.replace_entities(text)
        else:
            break
        n -= 1

    if n == 0:
        print(f'too many levels for: {repr(raw_text)}')
    return text


def get_pubmed_xml(pmid):
    """get pubmed xml with eutils
    """
    url = f'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=pubmed&id={pmid}&format=xml'
    return requests.get(url).text


def check_date(element):
    year = element.findtext('Year')
    month = element.findtext('Month')
    day = element.findtext('Day')

    if not month:
        date = parse_date(year).strftime('%Y')
    elif not day:
        date = parse_date(f'{year}-{month}').strftime('%Y-%m')
    else:
        date = parse_date(f'{year}-{month}-{day}').strftime('%Y-%m-%d')

    return date



import sys
import click

from pubmed_xml import version_info, Pubmed_XML_Parser, util


CONTEXT_SETTINGS = dict(help_option_names=['-?', '-h', '--help'])

__epilog__ = click.style('''

\b
example:
    {prog} --help
    {prog} 30003000
    {prog} 30003000,30003001,30003002
    {prog} 30003000 30003001 30003002
    {prog} 30003000,30003001,30003002 -o out.jl
    {prog} tests/pubmed22n1543.xml.gz -o out.jl

contact: {author} <{author_email}>
''', fg='cyan').format(**version_info)


@click.command(
    name=version_info['prog'],
    help=click.style(version_info['desc'], italic=True, fg='bright_green', bold=True),
    context_settings=CONTEXT_SETTINGS,
    epilog=__epilog__,
    no_args_is_help=True,
)
@click.argument('infile', nargs=-1)
@click.option('-o', '--outfile', help='the output filename')
@click.version_option(version=version_info['version'], prog_name=version_info['prog'])
def cli(**kwargs):
    outfile = kwargs['outfile']
    out = util.safe_open(outfile, 'w') if outfile else sys.stdout
    pubmed = Pubmed_XML_Parser()
    with out:
        for xml in kwargs['infile']:
            for article in pubmed.parse(xml):
                out.write(article.to_json() + '\n')
    if outfile:
        pubmed.logger.debug(f'save file to: {outfile}')


def main():
    cli()


if __name__ == '__main__':
    main()



import json
from pathlib import Path

from pubmed_xml.core.parser import Pubmed_XML_Parser


BASE_DIR = Path(__file__).resolve().parent
version_info = json.load(BASE_DIR.joinpath('version.json').open())

__version__ = version_info['version']



import json


class ArticleObject(object):

    def __init__(self, **context):
        self.data = context
        self.fields = list(context.keys())
        for k, v in context.items():
            setattr(self, k, v)

    def to_json(self, **kwargs):
        return json.dumps(self.data, ensure_ascii=False, **kwargs)

    def __getitem__(self, item):
        if item in self.data:
            return self.data[item]
        return f'no such attribute: {item}'

    def __str__(self):
        return f'Article<{self.pmid}>'

    __repr__ = __str__



from w3lib import html

from pubmed_xml import util


SPECIAL_CHARS = {
    u'\u2009': ' ',
    u'\u202f': ' ',
    u'\u00a0': ' ',
    u'\u2217': '*',
}


def parse_abstract(AbstractTexts):
    """
        - 1 no AbstractText
        - 2 sigle AbstractText
        - 3 multiple AbstractTexts with Labels
        
        * remove html tags
        * repalce html entities
        * repalce special chars
    """
    if not AbstractTexts:
        abstract = '.'
    elif len(AbstractTexts) == 1:
        abstract = ''.join(AbstractTexts[0].itertext()) or '.'
    else:
        abstracts = []
        for each in AbstractTexts:
            label = each.attrib.get('Label')
            text = ''.join(each.itertext())
            if label:
                abstracts.append('{}: {}'.format(label, text))
            else:
                abstracts.append(text)
        abstract = '\n'.join(abstracts)

    # ===========
    # remove tags
    # ===========
    abstract = html.remove_tags(abstract)

    # ===============
    # repalce entities
    # ===============
    abstract = util.replace_entities(abstract)

    # =====================
    # replace special chars
    # =====================
    for special, replace in SPECIAL_CHARS.items():
        abstract = abstract.replace(special, replace)

    return abstract



# -*- coding=utf-8 -*-
import os
import re
import datetime

try:
    import lxml.etree as ET
except ImportError:
    import xml.etree.cElementTree as ET

import click
from simple_loggers import SimpleLogger

from pubmed_xml import util
from pubmed_xml.core.article import ArticleObject
from pubmed_xml.core.abstract import parse_abstract


class Pubmed_XML_Parser(object):
    """PubMed XML Parser
    >>> from pubmed_xml import Pubmed_XML_Parser
    >>> pubmed = Pubmed_XML_Parser()
    >>> for article in pubmed.parse('30003000'):
    >>>     print(article)
    """

    def __init__(self):
        self.logger = SimpleLogger('Pubmed_XML_Parser')

    def get_tree(self, xml):
        if os.path.isfile(xml):
            tree = ET.parse(util.safe_open(xml))
        elif xml.startswith('<?xml '):
            tree = ET.fromstring(xml)
        else:
            tree = ET.fromstring(util.get_pubmed_xml(xml))
        return tree

    def parse(self, xml):
        """parse xml from local file or text string
        """
        try:
            tree = self.get_tree(xml)
        except Exception as e:
            raise Exception(click.style(f'[XML_PARSE_ERROR] {e}', fg='red'))

        if tree.find('PubmedArticle') is None:
            self.logger.warning('no article found')
            yield None
        else:
            for PubmedArticle in tree.iterfind('PubmedArticle'):
                context = {}
                MedlineCitation = PubmedArticle.find('MedlineCitation')
                Article = MedlineCitation.find('Article')

                context['pmid'] = int(MedlineCitation.findtext('PMID'))
                self.logger.debug('>>> PMID: {}'.format(context['pmid']))

                context['pub_mode'] = Article.attrib['PubModel']

                context['issn'] = Article.findtext('Journal/ISSN')
                context['issn_type'] = Article.find('Journal/ISSN').attrib['IssnType']

                context['journal'] = Article.findtext('Journal/Title')
                context['iso_abbr'] = Article.findtext('Journal/ISOAbbreviation')

                context['med_abbr'] = MedlineCitation.findtext('MedlineJournalInfo/MedlineTA')
                context['med_issn'] = MedlineCitation.findtext('MedlineJournalInfo/ISSNLinking')
                context['nlm_id'] = MedlineCitation.findtext('MedlineJournalInfo/NlmUniqueID')
                context['country'] = MedlineCitation.findtext('MedlineJournalInfo/Country')

                # PubDate vs ArticleDate vs PubMedPubDate?
                # https://gist.github.com/suqingdong/ad5166618b386627c0fea079215d77bb

                # convert to datetime
                context['mdat'] = util.check_date(PubmedArticle.find('MedlineCitation/DateRevised'))
                context['edat'] = util.check_date(PubmedArticle.find('PubmedData/History/PubMedPubDate[@PubStatus="pubmed"]'))
                context['pdat'] = util.check_date(Article.find('ArticleDate') if Article.find('ArticleDate') is not None else Article.find('Journal/JournalIssue/PubDate'))

                context['pagination'] = Article.findtext('Pagination/MedlinePgn')
                context['volume'] = Article.findtext('Journal/JournalIssue/Volume')
                context['issue'] = Article.findtext('Journal/JournalIssue/Issue')
                context['title'] = ''.join(Article.find('ArticleTitle').itertext())
                context['keywords'] = MedlineCitation.xpath('KeywordList/Keyword/text()')
                context['pub_status'] = PubmedArticle.findtext('PubmedData/PublicationStatus')

                context['abstract'] = parse_abstract(Article.xpath('Abstract/AbstractText'))
                
                author_mails = []
                author_list = []
                for author in Article.xpath('AuthorList/Author'):
                    names = [each.text for each in author.xpath('*')][:3]  # LastName, ForeName, Initials
                    names = [n for n in names if n]
                    fullname = names[0] if len(names) == 1 else ' '.join([names[1], names[0]])
                    author_list.append(fullname)     # fullname = ForeName LastName

                    affiliation = '\n'.join(author.xpath('AffiliationInfo/Affiliation/text()'))
                    mail = util.check_email(affiliation)
                    if mail:
                        author_mails.append(f'{fullname}:{mail}')
                context['author_mails'] = author_mails
                context['affiliations'] = list(set(Article.xpath('AuthorList/Author/AffiliationInfo/Affiliation/text()')))
                context['authors'] = author_list

                context['pub_types'] = Article.xpath('PublicationTypeList/PublicationType/text()')

                context['doi'] = PubmedArticle.findtext('PubmedData/ArticleIdList/ArticleId[@IdType="doi"]')
                context['pmc'] = PubmedArticle.findtext('PubmedData/ArticleIdList/ArticleId[@IdType="pmc"]')
                context['pii'] = PubmedArticle.findtext('PubmedData/ArticleIdList/ArticleId[@IdType="pii"]')

                references = PubmedArticle.xpath('PubmedData/ReferenceList/Reference/ArticleIdList/ArticleId[@IdType="pubmed"]/text()')
                context['references'] = ','.join(references)

                yield ArticleObject(**context)
